# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

bc_pcoa_results_url = ('https://data.qiime2.org/usage-examples/'
                       'moving-pictures/core-metrics-results/'
                       'bray_curtis_pcoa_results.qza')
unweighted_pcoa_results_url = ('https://data.qiime2.org/usage-examples/'
                               'moving-pictures/core-metrics-results/'
                               'unweighted_unifrac_pcoa_results.qza')

metadata_url = ('https://data.qiime2.org/usage-examples/'
                'moving-pictures/sample-metadata.tsv')


def plot(use):
    pcoa_results = use.init_artifact_from_url('pcoa_result',
                                              bc_pcoa_results_url)
    metadata = use.init_metadata_from_url('sample_metadata', metadata_url)

    viz, = use.action(
        use.UsageAction('emperor', 'plot'),
        use.UsageInputs(
            pcoa=pcoa_results,
            metadata=metadata
            ),
        use.UsageOutputNames(
            visualization='plot'
        )
    )

    viz.assert_output_type('Visualization')


def procrustes_plot(use):
    bc_pcoa_results = use.init_artifact_from_url('bc_pcoa_result',
                                                 bc_pcoa_results_url)
    unw_pcoa_results = use.init_artifact_from_url('unw_pcoa_result',
                                                  unweighted_pcoa_results_url)
    metadata = use.init_metadata_from_url('sample_metadata', metadata_url)

    viz, = use.action(
        use.UsageAction('emperor', 'procrustes_plot'),
        use.UsageInputs(
            reference_pcoa=bc_pcoa_results,
            other_pcoa=unw_pcoa_results,
            metadata=metadata
            ),
        use.UsageOutputNames(
            visualization='plot'
        )
    )

    viz.assert_output_type('Visualization')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


import q2_emperor
from ._plot import plot, procrustes_plot, biplot

from qiime2.plugin import (Plugin, Metadata, Str, List, Citations, Range, Int,
                           Bool, Properties)
from q2_types.ordination import PCoAResults, ProcrustesStatistics
import q2_emperor._examples as ex

PARAMETERS = {'metadata': Metadata, 'custom_axes': List[Str],
              'ignore_missing_samples': Bool}
PARAMETERS_DESC = {
    'metadata': 'The sample metadata.',
    'custom_axes': ('Numeric sample metadata columns that should be '
                    'included as axes in the Emperor plot.'),
    'ignore_missing_samples': (
        'This will suppress the error raised when the coordinates matrix '
        'contains samples that are not present in the metadata. Samples '
        'without metadata are included by setting all metadata values to: '
        '"This sample has no metadata". This flag is only applied if at '
        'least one sample is present in both the coordinates matrix and the '
        'metadata.'
    )
}

PLOT_PARAMETERS_DESC = {
    **PARAMETERS_DESC,
    'ignore_pcoa_features': 'Biplot arrows cannot be visualized using this '
                            'method. If you want to visualize biplot arrows '
                            'use the `biplot` method. Enabling this setting '
                            'will ignore any PCoA features that are present, '
                            'otherwise, if PCoA features are detected an '
                            'error will be raised.'
}

plugin = Plugin(
    name='emperor',
    version=q2_emperor.__version__,
    website='http://emperor.microbio.me',
    package='q2_emperor',
    citations=Citations.load('citations.bib', package='q2_emperor'),
    description=('This QIIME 2 plugin wraps Emperor and '
                 'supports interactive visualization of ordination '
                 'plots.'),
    short_description='Plugin for ordination plotting with Emperor.'
)

plugin.visualizers.register_function(
    function=plot,
    inputs={'pcoa': PCoAResults},
    parameters={
        'metadata': Metadata,
        'custom_axes': List[Str],
        'ignore_missing_samples': Bool,
        'ignore_pcoa_features': Bool
    },
    input_descriptions={
        'pcoa': 'The principal coordinates matrix to be plotted.'
    },
    parameter_descriptions=PLOT_PARAMETERS_DESC,
    name='Visualize and Interact with Principal Coordinates Analysis Plots',
    description='Generates an interactive ordination plot where the user '
                'can visually integrate sample metadata.',
    examples={'emperor_plot': ex.plot}
)

plugin.visualizers.register_function(
    function=procrustes_plot,
    inputs={'reference_pcoa': PCoAResults, 'other_pcoa': PCoAResults,
            'm2_stats': ProcrustesStatistics},
    parameters=PARAMETERS,
    input_descriptions={
        'reference_pcoa': 'The reference ordination matrix to be plotted.',
        'other_pcoa': 'The "other" ordination matrix to be plotted (the one '
                      'that was fitted to the reference).',
        'm2_stats': 'The M^2 value of the procrustes analysis & its associated'
                    ' p value.'
    },
    parameter_descriptions=PARAMETERS_DESC,
    name='Visualize and Interact with a procrustes plot',
    description='Plot two procrustes-fitted matrices',
    examples={'procrustes_plot': ex.procrustes_plot}
)

plugin.visualizers.register_function(
    function=biplot,
    inputs={'biplot': PCoAResults % Properties("biplot")},
    parameters={'sample_metadata': Metadata,
                'feature_metadata': Metadata,
                'ignore_missing_samples': Bool,
                'invert': Bool,
                'number_of_features': Int % Range(1, None)},
    input_descriptions={
        'biplot': 'The principal coordinates matrix to be plotted.'
    },
    parameter_descriptions={
        'sample_metadata': 'The sample metadata',
        'feature_metadata': 'The feature metadata (useful to manipulate the '
                            'arrows in the plot).',
        'invert': 'If specified, the point and arrow coordinates '
                  'will be swapped.',
        'ignore_missing_samples': PARAMETERS_DESC['ignore_missing_samples'],
        'number_of_features': 'The number of most important features '
                              '(arrows) to display in the ordination.'
                              ' “Importance” is calculated for each feature '
                              'based on the vector’s magnitude '
                              '(euclidean distance from origin).',
        },
    name='Visualize and Interact with Principal Coordinates Analysis Biplot',
    description='Generates an interactive ordination biplot where the user '
                'can visually integrate sample and feature metadata. '
                'Vectors representing the n most important features '
                'are then plotted in the emperor '
                'visualization (5 largest, by default).'
)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import tempfile
import unittest

import pandas as pd
import numpy as np
import qiime2
import skbio

from q2_emperor import plot, procrustes_plot, biplot, generic_plot

from qiime2.plugin.testing import TestPluginBase


class TestBase(TestPluginBase):
    package = 'q2_emperor.tests'

    def test_examples(self):
        self.execute_examples()


class PlotTests(unittest.TestCase):
    def setUp(self):
        eigvals = pd.Series(np.array([0.50, 0.25, 0.25]),
                            index=['PC1', 'PC2', 'PC3'])
        samples = np.array([[0.1, 0.2, 0.3],
                            [0.2, 0.3, 0.4],
                            [0.3, 0.4, 0.5],
                            [0.4, 0.5, 0.6]])
        proportion_explained = pd.Series([15.5, 12.2, 8.8],
                                         index=['PC1', 'PC2', 'PC3'])
        samples_df = pd.DataFrame(samples,
                                  index=['A', 'B', 'C', 'D'],
                                  columns=['PC1', 'PC2', 'PC3'])
        self.pcoa = skbio.OrdinationResults(
                'PCoA',
                'Principal Coordinate Analysis',
                eigvals,
                samples_df,
                proportion_explained=proportion_explained)

        samples_df = pd.DataFrame(samples + 1.01,
                                  index=['A', 'B', 'C', 'D'],
                                  columns=['PC1', 'PC2', 'PC3'])
        self.other = skbio.OrdinationResults(
                'PCoA',
                'Principal Coordinate Analysis',
                eigvals.copy(),
                samples_df,
                proportion_explained=proportion_explained.copy())

        features = pd.DataFrame(index=['x', 'y', 'z'],
                                columns=['PC1', 'PC2', 'PC3'],
                                data=[[1, 2, 3], [1, 1, -1], [0, -2, 0.1]])
        self.biplot = skbio.OrdinationResults(
                'PCoA',
                'Principal Coordinate Analysis',
                eigvals.copy(),
                samples_df,
                features=features,
                proportion_explained=proportion_explained.copy())

        self.metadata = qiime2.Metadata(
            pd.DataFrame({'val1': ['1.0', '2.0', '3.0', '4.0'],
                          'val2': ['3.3', '3.5', '3.6', '3.9']},
                         index=pd.Index(['A', 'B', 'C', 'D'], name='id')))

    def test_plot(self):
        with tempfile.TemporaryDirectory() as output_dir:
            plot(output_dir, self.pcoa, self.metadata)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_plot_custom_axis(self):
        with tempfile.TemporaryDirectory() as output_dir:
            plot(output_dir, self.pcoa, self.metadata, custom_axes=['val1'])
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_plot_custom_axes(self):
        with tempfile.TemporaryDirectory() as output_dir:
            plot(output_dir, self.pcoa, self.metadata,
                 custom_axes=['val1', 'val2'])
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_plot_with_pcoa_features_strategy_ignore(self):
        # using the biplot test data, since it has features
        with tempfile.TemporaryDirectory() as output_dir:
            plot(output_dir, self.biplot, self.metadata,
                 ignore_pcoa_features=True)
            self.assertTrue(True)  # checkpoint, everything should be good

    def test_plot_with_pcoa_features_strategy_error(self):
        # using the biplot test data, since it has features
        with tempfile.TemporaryDirectory() as output_dir, \
                self.assertRaisesRegex(ValueError,
                                       'Arrows cannot be visualized'):
            plot(output_dir, self.biplot, self.metadata,
                 ignore_pcoa_features=False)

    def test_plot_procrustes(self):
        with tempfile.TemporaryDirectory() as output_dir:
            procrustes_plot(output_dir, self.pcoa, other_pcoa=self.other,
                            metadata=self.metadata)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_plot_procrustes_custom_axis(self):
        with tempfile.TemporaryDirectory() as output_dir:
            procrustes_plot(output_dir, self.pcoa, other_pcoa=self.other,
                            metadata=self.metadata, custom_axes=['val1'])
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_generic_plot_with_settings(self):
        settings = {'shape': {'category': 'val1', 'data': {}}}
        with tempfile.TemporaryDirectory() as output_dir:
            generic_plot(output_dir, self.pcoa, self.metadata, other_pcoa=None,
                         plot_name='plot', settings=settings)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_biplot(self):
        with tempfile.TemporaryDirectory() as output_dir:
            biplot(output_dir, self.biplot,
                   sample_metadata=self.metadata)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_biplot_with_feature_metadata(self):
        feat_md = qiime2.Metadata(
            pd.DataFrame(index=pd.Index(['x', 'y', 'z'], name='feature id'),
                         columns=['k', 'p'],
                         data=[['Bacteria', 'Firmicutes'],
                               ['Bacteria', 'Firmicutes'],
                               ['Bacteria', 'Bacteroidetes']]))

        with tempfile.TemporaryDirectory() as output_dir:
            biplot(output_dir, self.biplot,
                   sample_metadata=self.metadata, feature_metadata=feat_md)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())

    def test_biplot_invert(self):
        feat_md = qiime2.Metadata(
            pd.DataFrame(index=pd.Index(['x', 'y', 'z'], name='feature id'),
                         columns=['k', 'p'],
                         data=[['Bacteria', 'Firmicutes'],
                               ['Bacteria', 'Firmicutes'],
                               ['Bacteria', 'Bacteroidetes']]))

        with tempfile.TemporaryDirectory() as output_dir:
            biplot(output_dir, self.biplot,
                   sample_metadata=self.metadata, feature_metadata=feat_md,
                   invert=True)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))
            self.assertTrue('src="./emperor.html"' in open(index_fp).read())



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest

from q2_emperor.plugin_setup import plugin as emperor_plugin


class PluginSetupTests(unittest.TestCase):

    def test_plugin_setup(self):
        self.assertEqual(emperor_plugin.name, 'emperor')



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import pkg_resources

import qiime2
import skbio
import q2templates
import math
import numpy as np
import pandas as pd
from emperor import Emperor
from scipy.spatial.distance import euclidean

TEMPLATES = pkg_resources.resource_filename('q2_emperor', 'assets')


def generic_plot(output_dir: str, master: skbio.OrdinationResults,
                 metadata: qiime2.Metadata,
                 other_pcoa: skbio.OrdinationResults,
                 plot_name: str, info: str = None,
                 custom_axes: str = None, settings: dict = None,
                 ignore_missing_samples: bool = False,
                 feature_metadata: qiime2.Metadata = None):

    mf = metadata.to_dataframe()
    if feature_metadata is not None:
        feature_metadata = feature_metadata.to_dataframe()

    if other_pcoa is None:
        procrustes = None
    else:
        procrustes = [other_pcoa]

    viz = Emperor(master, mf, feature_mapping_file=feature_metadata,
                  ignore_missing_samples=ignore_missing_samples,
                  procrustes=procrustes, remote='.')

    if custom_axes is not None:
        viz.custom_axes = custom_axes

    if other_pcoa:
        viz.procrustes_names = ['reference', 'other']

    viz.info = info
    viz.settings = settings

    html = viz.make_emperor(standalone=True)
    viz.copy_support_files(output_dir)
    with open(os.path.join(output_dir, 'emperor.html'), 'w') as fh:
        fh.write(html)

    index = os.path.join(TEMPLATES, 'index.html')
    q2templates.render(index, output_dir, context={'plot_name': plot_name})


def plot(output_dir: str, pcoa: skbio.OrdinationResults,
         metadata: qiime2.Metadata, custom_axes: str = None,
         ignore_missing_samples: bool = False,
         ignore_pcoa_features: bool = False) -> None:

    if ignore_pcoa_features:
        pcoa.features = None
    if pcoa.features is not None:
        raise ValueError("Arrows cannot be visualized with the 'plot' method, "
                         "use 'biplot' instead, or enable "
                         "`ignore_pcoa_features`.")

    generic_plot(output_dir, master=pcoa, metadata=metadata, other_pcoa=None,
                 ignore_missing_samples=ignore_missing_samples,
                 custom_axes=custom_axes, plot_name='plot')


def procrustes_plot(output_dir: str, reference_pcoa: skbio.OrdinationResults,
                    other_pcoa: skbio.OrdinationResults,
                    metadata: qiime2.Metadata, m2_stats: pd.DataFrame = None,
                    custom_axes: str = None,
                    ignore_missing_samples: bool = False) -> None:
    info = None
    if m2_stats is not None:
        m2 = '%.5f' % m2_stats['true M^2 value'][0]
        permutations = m2_stats['number of Monte Carlo permutations'][0]
        p_val = m2_stats['p-value for true M^2 value'][0]

        if permutations < 1:
            p_val = 'N/A'
        else:
            # Because the number of p-val dec places is dynamic, we need to
            # dynamically build the string template up.
            dec_places = math.ceil(math.log10(permutations))
            p_val_tmpl = '%%.%df' % dec_places
            p_val = p_val_tmpl % m2_stats['p-value for true M^2 value'][0]

        info = 'M&sup2; = %s p-value = %s' % (m2, p_val)

    generic_plot(output_dir, master=reference_pcoa, metadata=metadata,
                 other_pcoa=other_pcoa, custom_axes=custom_axes,
                 ignore_missing_samples=ignore_missing_samples,
                 plot_name='procrustes_plot', info=info)


def biplot(output_dir: str, biplot: skbio.OrdinationResults,
           sample_metadata: qiime2.Metadata, feature_metadata:
           qiime2.Metadata = None,
           ignore_missing_samples: bool = False,
           invert: bool = False,
           number_of_features: int = 5) -> None:

    if invert:
        biplot.samples, biplot.features = biplot.features, biplot.samples
        sample_metadata, feature_metadata = feature_metadata, sample_metadata

    # select the top N most important features based on the vector's magnitude
    feats = biplot.features.copy()
    origin = np.zeros_like(feats.columns)
    feats['importance'] = feats.apply(euclidean, axis=1, args=(origin,))
    feats.sort_values('importance', inplace=True, ascending=False)
    feats.drop(['importance'], inplace=True, axis=1)
    biplot.features = feats[:number_of_features].copy()

    generic_plot(output_dir, master=biplot, other_pcoa=None,
                 ignore_missing_samples=ignore_missing_samples,
                 metadata=sample_metadata, feature_metadata=feature_metadata,
                 plot_name='biplot')
